import pystokes.interface
import pystokes.periodic
import pystokes.twoWalls
import pystokes.unbounded
import pystokes.utils
import pystokes.wallBounded
import pystokes.forceFields
import pystokes.phoretic.unbounded
import pystokes.phoretic.wallBounded



## latest version of PyStokes
__version__="2.2.5"



cimport cython
from libc.math cimport sqrt
from cython.parallel import prange
cdef double PI = 3.14159265359
import numpy as np
cimport numpy as np


@cython.wraparound(False)
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
cdef class Rbm:
    """
    Rigid body motion (RBM) - velocity and angular velocity
    
    Methods in this class update velocities or angular velocities 
    using the inputs of -  arrays of positions, velocity or angular velocity, 
    along with an array of forces or torques or a slip mode

    The array of velocity or angular velocities is then update by each method. 
    
    ...

    ----------
    radius: float
        Radius of the particles (a).    
    particles: int
        Number of particles (Np)
    viscosity: float 
        Viscosity of the fluid (eta)

   """

    def __init__(self, radius=1, particles=1, viscosity=1.0):
        self.a   = radius
        self.Np  = particles
        self.eta = viscosity
        self.mu  = 1.0/(6*PI*self.eta*self.a)
        self.muv = 1.0/(8*PI*self.eta)
        self.mur = 1.0/(8*PI*self.eta*self.a**3)

        self.Mobility = np.zeros( (3*self.Np, 3*self.Np), dtype=np.float64)


    cpdef mobilityTT(self, double [:] v, double [:] r, double [:] F):
        """
        Compute velocity due to body forces using :math:`v=\mu^{TT}\cdot F` 
        ...

        Parameters
        ----------
        v: np.array
            An array of velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        F: np.array
            An array of forces
            An array of size 3*Np,
    
        Examples
        --------
        An example of the RBM 

        >>> import pystokes, numpy as np, matplotlib.pyplot as plt
        >>> # particle radius, self-propulsion speed, number and fluid viscosity
        >>> b, vs, Np, eta = 1.0, 1.0, 128, 0.1

        >>> #initialise
        >>> r = pystokes.utils.initialCondition(Np)  # initial random distribution of positions
        >>> p = np.zeros(3*Np); p[2*Np:3*Np] = -1    # initial orientation of the colloids
        >>> 
        >>> rbm   = pystokes.unbounded.Rbm(radius=b, particles=Np, viscosity=eta)
        >>> force = pystokes.forceFields.Forces(particles=Np)
        >>> 
        >>> def rhs(rp):
        >>>     # assign fresh values at each time step
        >>>     r = rp[0:3*Np];   p = rp[3*Np:6*Np]
        >>>     F, v, o = np.zeros(3*Np), np.zeros(3*Np), np.zeros(3*Np)
        >>> 
        >>>     force.lennardJonesWall(F, r, lje=0.01, ljr=5, wlje=1.2, wljr=3.4)
        >>>     rbm.mobilityTT(v, r, F)
        >>>     return np.concatenate( (v,o) )
        >>> 
        >>> # simulate the resulting system
        >>> Tf, Npts = 150, 200
        >>> pystokes.utils.simulate(np.concatenate((r,p)), 
        >>>      Tf,Npts,rhs,integrator='odeint', filename='crystallization')
        """


        cdef int Np  = self.Np, i, j, xx=2*Np
        cdef double dx, dy, dz, idr, idr2, vx, vy, vz, vv1, vv2, aa = (2.0*self.a*self.a)/3.0 
        cdef double mu=self.mu, muv=self.muv        
        
        for i in prange(Np, nogil=True):
            vx=0; vy=0;   vz=0;
            for j in range(Np):
                if i != j:
                    dx = r[i]    - r[j]
                    dy = r[i+Np] - r[j+Np]
                    dz = r[i+xx] - r[j+xx] 
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr2 = idr*idr
                    
                    vv1 = (1+aa*idr2)*idr 
                    vv2 = (1-3*aa*idr2)*( F[j]*dx + F[j+Np]*dy + F[j+xx]*dz )*idr2*idr
                    vx += vv1*F[j]    + vv2*dx 
                    vy += vv1*F[j+Np] + vv2*dy 
                    vz += vv1*F[j+xx] + vv2*dz 

            v[i]    += mu*F[i]    + muv*vx
            v[i+Np] += mu*F[i+Np] + muv*vy
            v[i+xx] += mu*F[i+xx] + muv*vz
        return 
               
   
    cpdef mobilityTR(self, double [:] v, double [:] r, double [:] T):
        """
        Compute velocity due to body torque using :math:`v=\mu^{TR}\cdot T` 
        ...

        Parameters
        ----------
        v: np.array
            An array of velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        T: np.array
            An array of torques
            An array of size 3*Np,
        """


        cdef int Np = self.Np, i, j, xx=2*Np 
        cdef double dx, dy, dz, idr, idr3, vx, vy, vz
        cdef double muv=self.muv       
        
        for i in prange(Np, nogil=True):
            vx=0; vy=0;   vz=0;
            for j in range(Np):
                if i != j:
                    dx = r[i]    - r[j]
                    dy = r[i+Np] - r[j+Np]
                    dz = r[i+xx] - r[j+xx] 
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr3 = idr*idr*idr
                    vx += (dy*T[j+xx] -T[j+Np]*dz )*idr3
                    vy += (dz*T[j]    -T[j+xx]*dx )*idr3
                    vz += (dx*T[j+Np] -T[j]   *dy )*idr3

            v[i]    += muv*vx
            v[i+Np] += muv*vy
            v[i+xx] += muv*vz
        return 
    
    
    cpdef propulsionT2s(self, double [:] v, double [:] r, double [:] S):
        """
        Compute velocity due to 2s mode of the slip :math:`v=\pi^{T,2s}\cdot S` 
        ...

        Parameters
        ----------
        v: np.array
            An array of velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        S: np.array
            An array of 2s mode of the slip
            An array of size 5*Np,
        """

        cdef int Np = self.Np, i, j, xx=2*Np, xx1=3*Np, xx2=4*Np
        cdef double dx, dy, dz, dr, idr,  idr3
        cdef double aa=(self.a*self.a*8.0)/3.0, vv1, vv2, aidr2
        cdef double vx, vy, vz, 
        cdef double sxx, sxy, sxz, syz, syy, srr, srx, sry, srz, mus = (28.0*self.a**3)/24 
 
        for i in prange(Np, nogil=True):
            vx=0; vy=0;   vz=0;
            for j in range(Np):
                if i != j:
                    sxx = S[j]
                    syy = S[j+Np]
                    sxy = S[j+xx]
                    sxz = S[j+xx1]
                    syz = S[j+xx2]
                    dx = r[i]    - r[j]
                    dy = r[i+Np] - r[j+Np]
                    dz = r[i+xx] - r[j+xx] 
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr3 = idr*idr*idr      
                    aidr2  = aa*idr*idr
                    
                    srr = (sxx*(dx*dx-dz*dz) + syy*(dy*dy-dz*dz) +  2*sxy*dx*dy + 2*sxz*dx*dz  +  2*syz*dy*dz)*idr*idr
                    srx = sxx*dx +  sxy*dy + sxz*dz  
                    sry = sxy*dx +  syy*dy + syz*dz  
                    srz = sxz*dx +  syz*dy - (sxx+syy)*dz 
                    
                    vv1 = 3*(1-aidr2)*srr*idr3
                    vv2 = 1.2*aidr2*idr3
                    vx +=  vv1*dx + vv2*srx
                    vy +=  vv1*dy + vv2*sry
                    vz +=  vv1*dz + vv2*srz
            
            v[i]   += vx*mus
            v[i+Np]+= vy*mus
            v[i+xx]+= vz*mus

        return 


    cpdef propulsionT3t(self, double [:] v, double [:] r, double [:] D):
        """
        Compute velocity due to 3t mode of the slip :math:`v=\pi^{T,3t}\cdot D` 
        ...

        Parameters
        ----------
        v: np.array
            An array of velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        D: np.array
            An array of 3t mode of the slip
            An array of size 3*Np,
        """

        cdef int Np = self.Np, i, j, xx=2*Np  
        cdef double dx, dy, dz, idr, idr3, Ddotidr, vx, vy, vz, mud = 3.0*self.a*self.a*self.a/5, mud1 = -1.0*(self.a**5)/10
 
        for i in prange(Np, nogil=True):
            vx=0; vy=0;   vz=0; 
            for j in range(Np):
                if i != j: 
                    dx = r[ i]   - r[j]
                    dy = r[i+Np] - r[j+Np]
                    dz = r[i+xx] - r[j+xx] 
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr3 = idr*idr*idr 
                    Ddotidr = (D[j]*dx + D[j+Np]*dy + D[j+xx]*dz)*idr*idr

                    vx += (D[j]    - 3.0*Ddotidr*dx )*idr3
                    vy += (D[j+Np] - 3.0*Ddotidr*dy )*idr3
                    vz += (D[j+xx] - 3.0*Ddotidr*dz )*idr3
            
            v[i]   += mud1*vx
            v[i+Np]+= mud1*vy
            v[i+xx]+= mud1*vz
        return 


    cpdef propulsionT3a(self, double [:] v, double [:] r, double [:] V):
        """
        Compute velocity due to 3a mode of the slip :math:`v=\pi^{T,3a}\cdot V` 
        ...

        Parameters
        ----------
        v: np.array
            An array of velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        V: np.array
            An array of 3a mode of the slip
            An array of size 5*Np,
        """

        cdef int Np = self.Np, i, j 
        cdef double dx, dy, dz, idr, idr5, vxx, vyy, vxy, vxz, vyz, vrx, vry, vrz
 
        for i in prange(Np, nogil=True):
            for j in range(Np):
                if i != j:
                    vxx = V[j]
                    vyy = V[j+Np]
                    vxy = V[j+2*Np]
                    vxz = V[j+3*Np]
                    vyz = V[j+4*Np]
                    dx = r[i]      - r[j]
                    dy = r[i+Np]   - r[j+Np]
                    dz = r[i+2*Np] - r[j+2*Np] 
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz)
                    idr5 = idr*idr*idr*idr*idr
                    vrx = vxx*dx +  vxy*dy + vxz*dz  
                    vry = vxy*dx +  vyy*dy + vyz*dz  
                    vrz = vxz*dx +  vyz*dy - (vxx+vyy)*dz 

                    v[i]      -= 8*( dy*vrz - dz*vry )*idr5
                    v[i+Np]   -= 8*( dz*vrx - dx*vrz )*idr5
                    v[i+2*Np] -= 8*( dx*vry - dy*vrx )*idr5 
                else:
                    pass
        return


    cpdef propulsionT3s(self, double [:] v, double [:] r, double [:] G):
        """
        Compute velocity due to 3s mode of the slip :math:`v=\pi^{T,3s}\cdot G` 
        ...

        Parameters
        ----------
        v: np.array
            An array of velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        G: np.array
            An array of 3s mode of the slip
            An array of size 7*Np,
        """

        cdef int Np = self.Np, i, j 
        cdef double dx, dy, dz, idr, idr5, idr7, aidr2, grrr, grrx, grry, grrz, gxxx, gyyy, gxxy, gxxz, gxyy, gxyz, gyyz
 
        for i in prange(Np, nogil=True):
             for j in range(Np):
                if i != j:
                    gxxx = G[j]
                    gyyy = G[j+Np]
                    gxxy = G[j+2*Np]
                    gxxz = G[j+3*Np]
                    gxyy = G[j+4*Np]
                    gxyz = G[j+5*Np]
                    gyyz = G[j+6*Np]
                    dx = r[i]      - r[j]
                    dy = r[i+Np]   - r[j+Np]
                    dz = r[i+2*Np] - r[j+2*Np] 
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr5 = idr*idr*idr*idr*idr      
                    idr7 = idr5*idr*idr     
                    aidr2 = (10.0/3)*self.a*self.a*idr*idr
                    
                    grrr = gxxx*dx*(dx*dx-3*dz*dz) + 3*gxxy*dy*(dx*dx-dz*dz) + gxxz*dz*(3*dx*dx-dz*dz) +\
                       3*gxyy*dx*(dy*dy-dz*dz) + 6*gxyz*dx*dy*dz + gyyy*dy*(dy*dy-3*dz*dz) +  gyyz*dz*(3*dy*dy-dz*dz) 
                    grrx = gxxx*(dx*dx-dz*dz) + gxyy*(dy*dy-dz*dz) +  2*gxxy*dx*dy + 2*gxxz*dx*dz  +  2*gxyz*dy*dz
                    grry = gxxy*(dx*dx-dz*dz) + gyyy*(dy*dy-dz*dz) +  2*gxyy*dx*dy + 2*gxyz*dx*dz  +  2*gyyz*dy*dz
                    grrz = gxxz*(dx*dx-dz*dz) + gyyz*(dy*dy-dz*dz) +  2*gxyz*dx*dy - 2*(gxxx+gxyy)*dx*dz  - 2*(gxxy+gyyy)*dy*dz
                  
                    v[i]      += 3*(1-(15.0/7)*aidr2)*grrx*idr5 - 15*(1-aidr2)*grrr*dx*idr7
                    v[i+Np]   += 3*(1-(15.0/7)*aidr2)*grry*idr5 - 15*(1-aidr2)*grrr*dy*idr7
                    v[i+2*Np] += 3*(1-(15.0/7)*aidr2)*grrz*idr5 - 15*(1-aidr2)*grrr*dz*idr7
                else:
                    pass 
        return


    cpdef propulsionT4a(self, double [:] v, double [:] r, double [:] M):
        """
        Compute velocity due to 4a mode of the slip :math:`v=\pi^{T,4a}\cdot M` 
        ...

        Parameters
        ----------
        v: np.array
            An array of velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        M: np.array
            An array of 4a mode of the slip
            An array of size 7*Np,
        """

        cdef int Np = self.Np, i, j 
        cdef double dx, dy, dz, idr, idr7
        cdef double mrrx, mrry, mrrz, mxxx, myyy, mxxy, mxxz, mxyy, mxyz, myyz
 
        for i in prange(Np, nogil=True):
            for j in range(Np):
                if i != j:
                    mxxx = M[j]
                    myyy = M[j+Np]
                    mxxy = M[j+2*Np]
                    mxxz = M[j+3*Np]
                    mxyy = M[j+4*Np]
                    mxyz = M[j+5*Np]
                    myyz = M[j+6*Np]
                    dx = r[i]      - r[j]
                    dy = r[i+Np]   - r[j+Np]
                    dz = r[i+2*Np] - r[j+2*Np] 
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr7 = idr*idr*idr*idr*idr*idr*idr
                    mrrx = mxxx*(dx*dx-dz*dz) + mxyy*(dy*dy-dz*dz) +  2*mxxy*dx*dy + 2*mxxz*dx*dz  +  2*mxyz*dy*dz
                    mrry = mxxy*(dx*dx-dz*dz) + myyy*(dy*dy-dz*dz) +  2*mxyy*dx*dy + 2*mxyz*dx*dz  +  2*myyz*dy*dz
                    mrrz = mxxz*(dx*dx-dz*dz) + myyz*(dy*dy-dz*dz) +  2*mxyz*dx*dy - 2*(mxxx+mxyy)*dx*dz  - 2*(mxxy+myyy)*dy*dz
                    
                    v[i]      -= 6*( dy*mrrz - dz*mrry )*idr7
                    v[i+Np]   -= 6*( dz*mrrx - dx*mrrz )*idr7
                    v[i+2*Np] -= 6*( dx*mrry - dy*mrrx )*idr7
                else:
                    pass
        return


    ## Angular velocities
    cpdef mobilityRT(self, double [:] o, double [:] r, double [:] F):
        """
        Compute angular velocity due to body forces using :math:`o=\mu^{RT}\cdot F` 
        ...

        Parameters
        ----------
        o: np.array
            An array of angular velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        F: np.array
            An array of forces
            An array of size 3*Np,
        """

        cdef int Np = self.Np, i, j, xx=2*Np 
        cdef double dx, dy, dz, idr, idr3, ox, oy, oz, muv=self.muv
 
        for i in prange(Np, nogil=True):
            ox=0;   oy=0;   oz=0;
            for j in range(Np):
                if i != j:
                    dx = r[i]    - r[j]
                    dy = r[i+Np] - r[j+Np]
                    dz = r[i+xx] - r[j+xx] 
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr3 = idr*idr*idr

                    ox += (F[j+Np]*dz - F[j+xx]*dy )*idr3
                    oy += (F[j+xx]*dx - F[j]   *dz )*idr3
                    oz += (F[j]   *dy - F[j+Np]*dx )*idr3
            o[i]    += muv*ox
            o[i+Np] += muv*oy
            o[i+xx] += muv*oz
        return  

               
    cpdef mobilityRR(   self, double [:] o, double [:] r, double [:] T):
        """
        Compute angular velocity due to body torques using :math:`o=\mu^{RR}\cdot T` 
        ...

        Parameters
        ----------
        o: np.array
            An array of angular velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        T: np.array
            An array of forces
            An array of size 3*Np,
        """

        cdef int Np = self.Np, i, j, xx=2*Np 
        cdef double dx, dy, dz, idr, idr3, Tdotidr, ox, oy, oz, mur=self.mur, muv=self.muv 
 
        for i in prange(Np, nogil=True):
            ox=0;   oy=0;   oz=0;
            for j in range(Np):
                if i != j:
                    dx = r[i]      - r[j]
                    dy = r[i+Np]   - r[j+Np]
                    dz = r[i+xx] - r[j+xx] 
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr3 = idr*idr*idr
                    Tdotidr = ( T[j]*dx + T[j+Np]*dy + T[j+xx]*dz )*idr*idr

                    ox += ( T[j]    - 3*Tdotidr*dx )*idr3
                    oy += ( T[j+Np] - 3*Tdotidr*dy )*idr3
                    oz += ( T[j+xx] - 3*Tdotidr*dz )*idr3
            
            o[i]    += mur*T[i]    - muv*ox
            o[i+Np] += mur*T[i+Np] - muv*oy
            o[i+xx] += mur*T[i+xx] - muv*oz
        return  

    
    cpdef propulsionR2s(self, double [:] o, double [:] r, double [:] S):
        """
        Compute angular velocity due to 2s mode of the slip :math:`v=\pi^{R,2s}\cdot S` 
        ...

        Parameters
        ----------
        o: np.array
            An array of angular velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        S: np.array
            An array of 2s mode of the slip
            An array of size 5*Np,
        """

        cdef int Np = self.Np, i, j, xx=2*Np 
        cdef double dx, dy, dz, idr, idr5, ox, oy, oz
        cdef double sxx, sxy, sxz, syz, syy, srr, srx, sry, srz, mus = (28.0*self.a*self.a*self.a)/24
 
        for i in prange(Np, nogil=True):
            ox=0;   oy=0;   oz=0;
            for j in range(Np):
                if i != j:
                    sxx = S[j]
                    syy = S[j+Np]
                    sxy = S[j+2*Np]
                    sxz = S[j+3*Np]
                    syz = S[j+4*Np]
                    dx = r[i]      - r[j]
                    dy = r[i+Np]   - r[j+Np]
                    dz = r[i+2*Np] - r[j+2*Np] 
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr5 = idr*idr*idr*idr*idr      
                    srx = sxx*dx +  sxy*dy + sxz*dz  
                    sry = sxy*dx +  syy*dy + syz*dz  
                    srz = sxz*dx +  syz*dy - (sxx+syy)*dz 

                    ox += 3*(sry*dz - srz*dy )*idr5
                    oy += 3*(srz*dx - srx*dz )*idr5
                    oz += 3*(srx*dy - sry*dx )*idr5
                    
            o[i]    += ox*mus
            o[i+Np] += oy*mus
            o[i+xx] += oz*mus
        return                 
    
    
    cpdef propulsionR3a(  self, double [:] o, double [:] r, double [:] V):
        """
        Compute angular velocity due to 3a mode of the slip :math:`v=\pi^{R,3a}\cdot V` 
        ...

        Parameters
        ----------
        o: np.array
            An array of angular velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        V: np.array
            An array of 3a mode of the slip
            An array of size 5*Np,
        """

        cdef int Np = self.Np, i, j 
        cdef double dx, dy, dz, idr, idr2, idr5, vxx, vyy, vxy, vxz, vyz, vrr, vrx, vry, vrz
 
        for i in prange(Np, nogil=True):
             for j in range(Np):
                if i != j:
                    vxx = V[j]
                    vyy = V[j+Np]
                    vxy = V[j+2*Np]
                    vxz = V[j+3*Np]
                    vyz = V[j+4*Np]
                    dx = r[i]      - r[j]
                    dy = r[i+Np]   - r[j+Np]
                    dz = r[i+2*Np] - r[j+2*Np] 
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr5 = idr*idr*idr*idr*idr      
                    vrr = (vxx*(dx*dx-dz*dz) + vyy*(dy*dy-dz*dz) +  2*vxy*dx*dy + 2*vxz*dx*dz  +  2*vyz*dy*dz)*idr*idr
                    vrx = vxx*dx +  vxy*dy + vxz*dz  
                    vry = vxy*dx +  vyy*dy + vyz*dz  
                    vrz = vxz*dx +  vyz*dy - (vxx+vyy)*dz 

                    o[i]      +=  ( 32*vrx- 20*vrr*dx )*idr5
                    o[i+Np]   +=  ( 32*vry- 20*vrr*dy )*idr5
                    o[i+2*Np] +=  ( 32*vrz- 20*vrr*dz )*idr5
                else :
                    pass 
        return


    cpdef propulsionR3s(  self, double [:] o, double [:] r, double [:] G):
        """
        Compute angular velocity due to 3s mode of the slip :math:`v=\pi^{R,3s}\cdot G` 
        ...

        Parameters
        ----------
        o: np.array
            An array of angular velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        G: np.array
            An array of 3s mode of the slip
            An array of size 7*Np,
        """


        cdef int Np = self.Np, i, j 
        cdef double dx, dy, dz, idr, idr7, grrx, grry, grrz, gxxx, gyyy, gxxy, gxxz, gxyy, gxyz, gyyz
 
        for i in prange(Np, nogil=True):
            for j in range(Np):
                if i != j:
                    gxxx = G[j]
                    gyyy = G[j+Np]
                    gxxy = G[j+2*Np]
                    gxxz = G[j+3*Np]
                    gxyy = G[j+4*Np]
                    gxyz = G[j+5*Np]
                    gyyz = G[j+6*Np]
                    dx = r[i]      - r[j]
                    dy = r[i+Np]   - r[j+Np]
                    dz = r[i+2*Np] - r[j+2*Np] 
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr7 = idr*idr*idr*idr*idr*idr*idr     
                    
                    grrx = gxxx*(dx*dx-dz*dz) + gxyy*(dy*dy-dz*dz) +  2*gxxy*dx*dy + 2*gxxz*dx*dz  +  2*gxyz*dy*dz
                    grry = gxxy*(dx*dx-dz*dz) + gyyy*(dy*dy-dz*dz) +  2*gxyy*dx*dy + 2*gxyz*dx*dz  +  2*gyyz*dy*dz
                    grrz = gxxz*(dx*dx-dz*dz) + gyyz*(dy*dy-dz*dz) +  2*gxyz*dx*dy - 2*(gxxx+gxyy)*dx*dz  - 2*(gxxy+gyyy)*dy*dz

                    o[i]      += 15*( dy*grrz - dz*grry )*idr7
                    o[i+Np]   += 15*( dz*grrx - dx*grrz )*idr7
                    o[i+2*Np] += 15*( dx*grry - dy*grrx )*idr7
                else :
                    pass
        return                 


    cpdef propulsionR4a(  self, double [:] o, double [:] r, double [:] M):
        """
        Compute angular velocity due to 4a mode of the slip :math:`v=\pi^{R,4a}\cdot M` 
        ...

        Parameters
        ----------
        o: np.array
            An array of angular velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        M: np.array
            An array of 4a mode of the slip
            An array of size 7*Np,
        """


        cdef int Np = self.Np, i, j 
        cdef double dx, dy, dz, idr, idr7, idr9, mrrr, mrrx, mrry, mrrz, mxxx, myyy, mxxy, mxxz, mxyy, mxyz, myyz
 
        for i in prange(Np, nogil=True):
             for j in range(Np):
                if i != j:
                    mxxx = M[j]
                    myyy = M[j+Np]
                    mxxy = M[j+2*Np]
                    mxxz = M[j+3*Np]
                    mxyy = M[j+4*Np]
                    mxyz = M[j+5*Np]
                    myyz = M[j+6*Np]
                    dx = r[i]      - r[j]
                    dy = r[i+Np]   - r[j+Np]
                    dz = r[i+2*Np] - r[j+2*Np] 
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr7 = idr*idr*idr*idr*idr*idr*idr      
                    idr9 = idr7*idr*idr     
                    
                    mrrr = mxxx*dx*(dx*dx-3*dz*dz) + 3*mxxy*dy*(dx*dx-dz*dz) + mxxz*dz*(3*dx*dx-dz*dz) +\
                       3*mxyy*dx*(dy*dy-dz*dz) + 6*mxyz*dx*dy*dz + myyy*dy*(dy*dy-3*dz*dz) +  myyz*dz*(3*dy*dy-dz*dz) 
                    mrrx = mxxx*(dx*dx-dz*dz) + mxyy*(dy*dy-dz*dz) +  2*mxxy*dx*dy + 2*mxxz*dx*dz  +  2*mxyz*dy*dz
                    mrry = mxxy*(dx*dx-dz*dz) + myyy*(dy*dy-dz*dz) +  2*mxyy*dx*dy + 2*mxyz*dx*dz  +  2*myyz*dy*dz
                    mrrz = mxxz*(dx*dx-dz*dz) + myyz*(dy*dy-dz*dz) +  2*mxyz*dx*dy - 2*(mxxx+mxyy)*dx*dz  - 2*(mxxy+myyy)*dy*dz
                  
                    o[i]      += 21*mrrr*dx*idr9 - 9*mrrx*idr7  
                    o[i+Np]   += 21*mrrr*dy*idr9 - 9*mrry*idr7  
                    o[i+2*Np] += 21*mrrr*dz*idr9 - 9*mrrz*idr7  
                else:
                    pass 
        return


    cpdef noiseTT(self, double [:] v, double [:] r):
        """
        Compute translation Brownian motion 
        ...

        Parameters
        ----------
        v: np.array
            An array of velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        """

        cdef int i, j, Np=self.Np, xx=2*Np
        cdef double dx, dy, dz, idr, h2, hsq, idr2, idr3, idr4, idr5
        cdef double mu=self.mu, muv=2*mu*self.a*0.75, a2=self.a*self.a/3.0
        cdef double vx, vy, vz, mm=1/(.75*self.a)

        cdef double [:, :] M = self.Mobility
        cdef double [:]    Fr = np.random.normal(size=3*Np)


        for i in prange(Np, nogil=True):
            for j in range(Np):
                dx = r[i]    - r[j]
                dy = r[i+Np] - r[j+Np]
                h2=2*r[j+xx]; hsq=r[j+xx]*r[j+xx]
                if i!=j:
                    dz = r[i+xx] - r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr2=idr*idr;  idr3=idr*idr*idr
                    dx = dx*idr; dy=dy*idr; dz=dz*idr
                    #
                    M[i,    j   ] = (1 + dx*dx)*idr + a2*(2 - 6*dx*dx)*idr3
                    M[i+Np, j+Np] = (1 + dy*dy)*idr + a2*(2 - 6*dy*dy)*idr3
                    M[i+xx, j+xx] = (1 + dz*dz)*idr + a2*(2 - 6*dz*dz)*idr3
                    M[i,    j+Np] = (    dx*dy)*idr + a2*(  - 6*dx*dy)*idr3
                    M[i,    j+xx] = (    dx*dz)*idr + a2*(  - 6*dx*dz)*idr3
                    M[i+Np, j+xx] = (    dy*dz)*idr + a2*(  - 6*dy*dz)*idr3
                else:
                    # one-body mobility
                    M[i,    j   ] = mm
                    M[i+Np, j+Np] = mm
                    M[i+xx, j+xx] = mm
                    M[i,    j+Np] = 0
                    M[i,    j+xx] = 0
                    M[i+Np, j+xx] = 0


        for i in prange(Np, nogil=True):
            for j in range(Np):
                M[i,    j   ] = muv*M[i,    j   ]
                M[i+Np, j+Np] = muv*M[i+Np, j+Np]
                M[i+xx, j+xx] = muv*M[i+xx, j+xx]
                M[i,    j+Np] = muv*M[i,    j+Np]
                M[i,    j+xx] = muv*M[i,    j+xx]
                M[i+Np, j+xx] = muv*M[i+Np, j+xx]

                M[i+Np, j   ] =     M[i,    j+Np]
                M[i+xx, j   ] =     M[i,    j+xx]
                M[i+xx, j+Np] =     M[i+Np, j+xx]

        cdef double [:, :] L = np.linalg.cholesky(self.Mobility)

        for i in prange(Np, nogil=True):
            vx=0; vy=0; vz=0;
            for j in range(Np):
                vx += L[i   , j]*Fr[j] + L[i   , j+Np]*Fr[j+Np] + L[i   , j+xx]*Fr[j+xx]
                vy += L[i+Np, j]*Fr[j] + L[i+Np, j+Np]*Fr[j+Np] + L[i+Np, j+xx]*Fr[j+xx]
                vz += L[i+xx, j]*Fr[j] + L[i+xx, j+Np]*Fr[j+Np] + L[i+xx, j+xx]*Fr[j+xx]
            v[i  ]  += vx
            v[i+Np] += vy
            v[i+xx] += vz

        return


    cpdef noiseRR(self, double [:] o, double [:] r):
        """
        Compute rotational Brownian motion 
        ...

        Parameters
        ----------
        o: np.array
            An array of angular velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        """

        cdef int i, j, Np=self.Np, xx=2*Np
        cdef double dx, dy, dz, idr, h2, hsq, idr2, idr3, idr4, idr5
        cdef double mur=1/(8*np.pi*self.eta), muv=0.25*sqrt(2.0)*mur, mm=4/(self.a**3)
        cdef double ox, oy, oz

        cdef double [:, :] M = self.Mobility
        cdef double [:]   Tr = np.random.normal(size=3*Np)


        for i in prange(Np, nogil=True):
            for j in range(Np):
                dx = r[i]    - r[j]
                dy = r[i+Np] - r[j+Np]
                h2=2*r[j+xx]; hsq=r[j+xx]*r[j+xx]
                if i!=j:
                    dz = r[i+xx] - r[j+xx]
                    idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                    idr2=idr*idr;  idr3=idr*idr*idr
                    dx = dx*idr; dy=dy*idr; dz=dz*idr
                    #
                    M[i,    j   ] = (2 - 6*dx*dx)*idr3
                    M[i+Np, j+Np] = (2 - 6*dy*dy)*idr3
                    M[i+xx, j+xx] = (2 - 6*dz*dz)*idr3
                    M[i,    j+Np] = (  - 6*dx*dy)*idr3
                    M[i,    j+xx] = (  - 6*dx*dz)*idr3
                    M[i+Np, j+xx] = (  - 6*dy*dz)*idr3


        for i in prange(Np, nogil=True):
            for j in range(Np):
                M[i,    j   ] = muv*M[i,    j   ]
                M[i+Np, j+Np] = muv*M[i+Np, j+Np]
                M[i+xx, j+xx] = muv*M[i+xx, j+xx]
                M[i,    j+Np] = muv*M[i,    j+Np]
                M[i,    j+xx] = muv*M[i,    j+xx]
                M[i+Np, j+xx] = muv*M[i+Np, j+xx]

                M[i+Np, j   ] =     M[i,    j+Np]
                M[i+xx, j   ] =     M[i,    j+xx]
                M[i+xx, j+Np] =     M[i+Np, j+xx]

        cdef double [:, :] L = muv*np.linalg.cholesky(self.Mobility)
        for i in prange(Np, nogil=True):
            ox=0; oy=0; oz=0;
            for j in range(Np):
                ox += L[i   , j]*Tr[j] + L[i   , j+Np]*Tr[j+Np] + L[i   , j+xx]*Tr[j+xx]
                oy += L[i+Np, j]*Tr[j] + L[i+Np, j+Np]*Tr[j+Np] + L[i+Np, j+xx]*Tr[j+xx]
                oz += L[i+xx, j]*Tr[j] + L[i+xx, j+Np]*Tr[j+Np] + L[i+xx, j+xx]*Tr[j+xx]
            o[i  ]  += ox
            o[i+Np] += oy
            o[i+xx] += oz
        return



## Flow at given points
@cython.wraparound(False)
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
cdef class Flow:
    """
    Flow at given points
    
    ...

    Parameters
    ----------
    radius: float
        Radius of the particles.    
    particles: int
        Number of particles 
    viscosity: viscosity of the fluid 
    gridpoints: int 
        Number of grid points
    Examples
    --------
    An example of the RBM

    """


    def __init__(self, radius=1, particles=1, viscosity=1, gridpoints=32):
        self.a  = radius
        self.Np = particles
        self.Nt = gridpoints
        self.eta= viscosity

    cpdef flowField1s(self, double [:] vv, double [:] rt, double [:] r, double [:] F):
        """
        Compute flow field at field points due body forces
        ...

        Parameters
        ----------
        vv: np.array
            An array of flow at field points
            An array of size 3*Nt,
        rt: np.array
            An array of field points
            An array of size 3*Nt,
        r: np.array
            An array of positions
            An array of size 3*Np,
        F: np.array
            An array of body force
            An array of size 3*Np,
    
        Examples
        --------
        An example of the Flow field due to $1s$ mode of force per unit area

        >>> import pystokes, numpy as np, matplotlib.pyplot as plt
        >>> 
        >>> # particle radius, self-propulsion speed, number and fluid viscosity
        >>> b, eta, Np = 1.0, 1.0/6.0, 1
        >>> 
        >>> # initialize
        >>> r, p = np.array([0.0, 0.0, 3.4]), np.array([0.0, 1.0, 0])
        >>> F1s  = pystokes.utils.irreducibleTensors(1, p)
        >>> 
        >>> # space dimension , extent , discretization
        >>> dim, L, Ng = 3, 10, 64;
        >>> 
        >>> # instantiate the Flow class
        >>> flow = pystokes.unbounded.Flow(radius=b, particles=Np, viscosity=eta, gridpoints=Ng*Ng)
        >>> 
        >>> # create grid, evaluate flow and plot
        >>> rr, vv = pystokes.utils.gridXY(dim, L, Ng)
        >>> flow.flowField1s(vv, rr, r, F1s)
        >>> pystokes.utils.plotStreamlinesXY(vv, rr, r, offset=6-1, density=1.4, title='1s')
        """

        cdef int Np = self.Np,  Nt = self.Nt
        cdef int i, ii, xx = 2*Np
        cdef double dx, dy, dz, idr, idr2, vv1, vv2, vx, vy, vz, muv = 1/(8*PI*self.eta), aa = self.a*self.a/3.0
        for i in prange(Nt, nogil=True):
            vx = 0.0; vy = 0.0; vz = 0.0;
            for ii in range(Np):
                
                dx = rt[i]      - r[ii]
                dy = rt[i+Nt]   - r[ii+Np]
                dz = rt[i+2*Nt] - r[ii+xx] 
                idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                idr2= idr*idr
                vv1 = (1+aa*idr2)*idr 
                vv2 = (1-3*aa*idr2)*( F[ii]*dx + F[ii+Np]*dy + F[ii+xx]*dz )*idr2*idr
                
                vx += vv1*F[ii]     + vv2*dx
                vy += vv1*F[ii+ Np] + vv2*dy
                vz += vv1*F[ii+ xx] + vv2*dz
            
            vv[i]         += vx*muv
            vv[i +   Nt]  += vy*muv
            vv[i +  2*Nt] += vz*muv
        return 


    cpdef flowField2a(self, double [:] vv, double [:] rt, double [:] r, double [:] T):
        """
        Compute flow field at field points due body Torque 
        ...

        Parameters
        ----------
        vv: np.array
            An array of flow at field points
            An array of size 3*Nt,
        rt: np.array
            An array of field points
            An array of size 3*Nt,
        r: np.array
            An array of positions
            An array of size 3*Np,
        T: np.array
            An array of body torque
            An array of size 3*Np,
    
        Examples
        --------
        An example of the RBM 

        # Example 1: Flow field due to $2a$ mode of force per unit area
        >>> import pystokes, numpy as np, matplotlib.pyplot as plt
        >>> 
        >>> # particle radius, self-propulsion speed, number and fluid viscosity
        >>> b, eta, Np = 1.0, 1.0/6.0, 1
        >>> 
        >>> # initialize
        >>> r, p = np.array([0.0, 0.0, 3.4]), np.array([0.0, 1.0, 0])
        >>> V2a  = pystokes.utils.irreducibleTensors(1, p)
        >>> 
        >>> # space dimension , extent , discretization
        >>> dim, L, Ng = 3, 10, 64;
        >>> 
        >>> # instantiate the Flow class
        >>> flow = pystokes.wallBounded.Flow(radius=b, particles=Np, viscosity=eta, gridpoints=Ng*Ng)
        >>> 
        >>> # create grid, evaluate flow and plot
        >>> rr, vv = pystokes.utils.gridXY(dim, L, Ng)
        >>> flow.flowField2a(vv, rr, r, V2a)
        >>> pystokes.utils.plotStreamlinesXY(vv, rr, r, offset=6-1, density=1.4, title='2s')
        """

        cdef int Np = self.Np, Nt = self.Nt
        cdef int i, ii, xx = 2*Np
        cdef double dx, dy, dz, idr, idr3, vx, vy, vz, mur1 = 1.0/(8*PI*self.eta)
        for i in prange(Nt, nogil=True):
            vx = 0.0; vy = 0.0; vz = 0.0;
            for ii in range(Np):
                dx = rt[i]      - r[ii]
                dy = rt[i+Nt]   - r[ii+Np]
                dz = rt[i+2*Nt] - r[ii+xx] 
                idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                idr3 = idr*idr*idr
          
                vx += ( dy*T[ii+xx] - dz*T[ii+Np])*idr3
                vy += ( dz*T[ii]    - dx*T[ii+xx])*idr3 
                vz += ( dx*T[ii+Np] - dy*T[ii]   )*idr3

            vv[i  ]      += vx*mur1
            vv[i + Nt]   += vy*mur1
            vv[i + 2*Nt] += vz*mur1
        return  


    cpdef flowField2s(self, double [:] vv, double [:] rt, double [:] r, double [:] S):
        """
        Compute flow field at field points  due to 2s mode of the slip 
        ...

        Parameters
        ----------
        vv: np.array
            An array of flow at field points
            An array of size 3*Nt,
        rt: np.array
            An array of field points
            An array of size 3*Nt,
        r: np.array
            An array of positions
            An array of size 3*Np,
        S: np.array
            An array of 2s mode of the slip
            An array of size 5*Np,
        
        Examples
        --------
        An example of the Flow field due to $3t$ mode of active slip

        >>> import pystokes, numpy as np, matplotlib.pyplot as plt
        >>> 
        >>> # particle radius, self-propulsion speed, number and fluid viscosity
        >>> b, eta, Np = 1.0, 1.0/6.0, 1
        >>> 
        >>> # initialize
        >>> r, p = np.array([0.0, 0.0, 3.4]), np.array([0.0, 1.0, 0])
        >>> V3t  = pystokes.utils.irreducibleTensors(1, p)
        >>> 
        >>> # space dimension , extent , discretization
        >>> dim, L, Ng = 3, 10, 64;
        >>> 
        >>> # instantiate the Flow class
        >>> flow = pystokes.wallBounded.Flow(radius=b, particles=Np, viscosity=eta, gridpoints=Ng*Ng)
        >>> 
        >>> # create grid, evaluate flow and plot
        >>> rr, vv = pystokes.utils.gridXY(dim, L, Ng)
        >>> flow.flowField3t(vv, rr, r, V3t)
        >>> pystokes.utils.plotStreamlinesXY(vv, rr, r, offset=6-1, density=1.4, title='1s')
        """

        cdef int Np = self.Np, Nt = self.Nt
        cdef int i, ii, xx= 2*Np, xx1= 3*Np, xx2 = 4*Np
        cdef double dx, dy, dz, idr, idr3, aidr2, sxx, syy, sxy, sxz, syz, srr, srx, sry, srz
        cdef double aa = self.a**2, vv1, vv2, vx, vy, vz, mus = (28.0*self.a**3)/24 
        for i in prange(Nt, nogil=True):
            vx = 0.0;vy = 0.0; vz = 0.0;
            for ii in range(Np):
                sxx = S[ii]
                syy = S[ii+Np]
                sxy = S[ii+xx]
                sxz = S[ii+xx1]
                syz = S[ii+xx2]
                
                dx = rt[i]      - r[ii]
                dy = rt[i+Nt]   - r[ii+Np]
                dz = rt[i+2*Nt] - r[ii+xx] 
                
                idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                idr3 = idr*idr*idr      
                aidr2  = aa*idr*idr
                
                srr = (sxx*(dx*dx-dz*dz) + syy*(dy*dy-dz*dz) +  2*sxy*dx*dy + 2*sxz*dx*dz  +  2*syz*dy*dz)*idr*idr
                srx = sxx*dx +  sxy*dy + sxz*dz  
                sry = sxy*dx +  syy*dy + syz*dz  
                srz = sxz*dx +  syz*dy - (sxx+syy)*dz 
                
                vv1 = 3*(1-aidr2)*srr*idr3
                vv2 = 1.2*aidr2*idr3
                
                vx +=  vv1*dx + vv2*srx
                vy +=  vv1*dy + vv2*sry
                vz +=  vv1*dz + vv2*srz
                 
            vv[i]      +=  vx*mus
            vv[i+Nt]   +=  vy*mus
            vv[i+2*Nt] +=  vz*mus
        return


    cpdef flowField3t(self, double [:] vv, double [:] rt, double [:] r, double [:] D):
        """
        Compute flow field at field points due to 3t mode of the slip 
        ...

        Parameters
        ----------
        vv: np.array
            An array of flow at field points
            An array of size 3*Nt,
        rt: np.array
            An array of field points
            An array of size 3*Nt,
        r: np.array
            An array of positions
            An array of size 3*Np,
        D: np.array
            An array of 3t mode of the slip
            An array of size 3*Np,
 
        Examples
        --------
        An example of the Flow field due to $3t$ mode of active slip

        >>> import pystokes, numpy as np, matplotlib.pyplot as plt
        >>> 
        >>> # particle radius, self-propulsion speed, number and fluid viscosity
        >>> b, eta, Np = 1.0, 1.0/6.0, 1
        >>> 
        >>> # initialize
        >>> r, p = np.array([0.0, 0.0, 3.4]), np.array([0.0, 1.0, 0])
        >>> V3t  = pystokes.utils.irreducibleTensors(1, p)
        >>> 
        >>> # space dimension , extent , discretization
        >>> dim, L, Ng = 3, 10, 64;
        >>> 
        >>> # instantiate the Flow class
        >>> flow = pystokes.wallBounded.Flow(radius=b, particles=Np, viscosity=eta, gridpoints=Ng*Ng)
        >>> 
        >>> # create grid, evaluate flow and plot
        >>> rr, vv = pystokes.utils.gridXY(dim, L, Ng)
        >>> flow.flowField3t(vv, rr, r, V3t)
        >>> pystokes.utils.plotStreamlinesXY(vv, rr, r, offset=6-1, density=1.4, title='2s')
        """

        cdef int Np = self.Np, Nt = self.Nt
        cdef  int i, ii 
        cdef double dx, dy, dz, idr, idr3, Ddotidr, vx, vy, vz,mud1 = -1.0*(self.a**5)/10
 
        for i in prange(Nt, nogil=True):
            vx =0.0; vy = 0.0; vz =0.0;
            for ii in range(Np):
                dx = rt[i]      - r[ii]
                dy = rt[i+Nt]   - r[ii+Np]
                dz = rt[i+2*Nt] - r[ii+2*Np] 
                idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                idr3 = idr*idr*idr 
                Ddotidr = (D[ii]*dx + D[ii+Np]*dy + D[ii+2*Np]*dz)*idr*idr

                vx += (D[ii]      - 3.0*Ddotidr*dx )*idr3
                vy += (D[ii+Np]   - 3.0*Ddotidr*dy )*idr3
                vz += (D[ii+2*Np] - 3.0*Ddotidr*dz )*idr3
        
            vv[i]      += vx*mud1
            vv[i+Nt]   += vy*mud1
            vv[i+2*Nt] += vz*mud1
        
        return 


    cpdef flowField3s(self, double [:] vv, double [:] rt, double [:] r, double [:] G):
        """
        Compute flow field at field points  due to 3s mode of the slip 
        ...

        Parameters
        ----------
        vv: np.array
            An array of flow at field points
            An array of size 3*Nt,
        rt: np.array
            An array of field points
            An array of size 3*Nt,
        r: np.array
            An array of positions
            An array of size 3*Np,
        G: np.array
            An array of 3s mode of the slip
            An array of size 7*Np,
        """

        cdef int Np = self.Np, Nt = self.Nt
        cdef int i, ii, 
        cdef double dx, dy, dz, idr, idr5, idr7
        cdef double aidr2, grrr, grrx, grry, grrz, gxxx, gyyy, gxxy, gxxz, gxyy, gxyz, gyyz
        for i in prange(Nt, nogil=True):
            for ii in range(Np):
                gxxx = G[ii]
                gyyy = G[ii+Np]
                gxxy = G[ii+2*Np]
                gxxz = G[ii+3*Np]
                gxyy = G[ii+4*Np]
                gxyz = G[ii+5*Np]
                gyyz = G[ii+6*Np]
                dx = rt[i]      - r[ii]
                dy = rt[i+Nt]   - r[ii+Np]
                dz = rt[i+2*Nt] - r[ii+2*Np] 
                idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                idr5 = idr*idr*idr*idr*idr      
                idr7 = idr5*idr*idr     
                aidr2 = self.a*self.a*idr*idr
                
                grrr = gxxx*dx*(dx*dx-3*dz*dz) + 3*gxxy*dy*(dx*dx-dz*dz) + gxxz*dz*(3*dx*dx-dz*dz) +\
                       3*gxyy*dx*(dy*dy-dz*dz) + 6*gxyz*dx*dy*dz + gyyy*dy*(dy*dy-3*dz*dz) +  gyyz*dz*(3*dy*dy-dz*dz) 
                grrx = gxxx*(dx*dx-dz*dz) + gxyy*(dy*dy-dz*dz) +  2*gxxy*dx*dy + 2*gxxz*dx*dz  +  2*gxyz*dy*dz
                grry = gxxy*(dx*dx-dz*dz) + gyyy*(dy*dy-dz*dz) +  2*gxyy*dx*dy + 2*gxyz*dx*dz  +  2*gyyz*dy*dz
                grrz = gxxz*(dx*dx-dz*dz) + gyyz*(dy*dy-dz*dz) +  2*gxyz*dx*dy - 2*(gxxx+gxyy)*dx*dz  - 2*(gxxy+gyyy)*dy*dz
             
                vv[i]      += 3*(1-(15.0/7)*aidr2)*grrx*idr5 - 15*(1-aidr2)*grrr*dx*idr7
                vv[i+Nt]   += 3*(1-(15.0/7)*aidr2)*grry*idr5 - 15*(1-aidr2)*grrr*dy*idr7
                vv[i+2*Nt] += 3*(1-(15.0/7)*aidr2)*grrz*idr5 - 15*(1-aidr2)*grrr*dz*idr7
        return


    cpdef flowField3a(self, double [:] vv, double [:] rt, double [:] r, double [:] V):
        """
        Compute flow field at field points  due to 3a mode of the slip 
        ...

        Parameters
        ----------
        vv: np.array
            An array of flow at field points
            An array of size 3*Nt,
        rt: np.array
            An array of field points
            An array of size 3*Nt,
        r: np.array
            An array of positions
            An array of size 3*Np,
        V: np.array
            An array of 3a mode of the slip
            An array of size 5*Np,
        """

        cdef int Np = self.Np, Nt = self.Nt
        cdef int i, ii
        cdef double dx, dy, dz, idr, idr5, vxx, vyy, vxy, vxz, vyz, vrx, vry, vrz
 
        for i in prange(Nt, nogil=True):
            for ii in range(Np):
                vxx = V[ii]
                vyy = V[ii+Np]
                vxy = V[ii+2*Np]
                vxz = V[ii+3*Np]
                vyz = V[ii+4*Np]
                dx = rt[i]      - r[ii]
                dy = rt[i+Nt]   - r[ii+Np]
                dz = rt[i+2*Nt] - r[ii+2*Np] 
                idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz)
                idr5 = idr*idr*idr*idr*idr
                vrx = vxx*dx +  vxy*dy + vxz*dz  
                vry = vxy*dx +  vyy*dy + vyz*dz  
                vrz = vxz*dx +  vyz*dy - (vxx+vyy)*dz 

                vv[i]      += 8*( dy*vrz - dz*vry )*idr5
                vv[i+Nt]   += 8*( dz*vrx - dx*vrz )*idr5
                vv[i+2*Nt] += 8*( dx*vry - dy*vrx )*idr5 
        return 


    cpdef flowField4a(self, double [:] vv, double [:] rt, double [:] r, double [:] M):
        """
        Compute flow field at field points  due to 4a mode of the slip 
        ...

        Parameters
        ----------
        vv: np.array
            An array of flow at field points
            An array of size 3*Nt,
        rt: np.array
            An array of field points
            An array of size 3*Nt,
        r: np.array
            An array of positions
            An array of size 3*Np,
        M: np.array
            An array of 4a mode of the slip
            An array of size 7*Np,
        """

        cdef int Np = self.Np, Nt = self.Nt
        cdef int i, ii
        cdef double dx, dy, dz, idr, idr7
        cdef double mrrx, mrry, mrrz, mxxx, myyy, mxxy, mxxz, mxyy, mxyz, myyz
 
        for i in prange(Nt, nogil=True):
            for ii in range(Np):
                mxxx = M[ii]
                myyy = M[ii+Np]
                mxxy = M[ii+2*Np]
                mxxz = M[ii+3*Np]
                mxyy = M[ii+4*Np]
                mxyz = M[ii+5*Np]
                myyz = M[ii+6*Np]
                dx = rt[i]      - r[ii]
                dy = rt[i+Nt]   - r[ii+Np]
                dz = rt[i+2*Nt] - r[ii+2*Np] 
                idr = 1.0/sqrt( dx*dx + dy*dy + dz*dz )
                idr7 = idr*idr*idr*idr*idr*idr*idr
                mrrx = mxxx*(dx*dx-dz*dz) + mxyy*(dy*dy-dz*dz) +  2*mxxy*dx*dy + 2*mxxz*dx*dz  +  2*mxyz*dy*dz
                mrry = mxxy*(dx*dx-dz*dz) + myyy*(dy*dy-dz*dz) +  2*mxyy*dx*dy + 2*mxyz*dx*dz  +  2*myyz*dy*dz
                mrrz = mxxz*(dx*dx-dz*dz) + myyz*(dy*dy-dz*dz) +  2*mxyz*dx*dy - 2*(mxxx+mxyy)*dx*dz  - 2*(mxxy+myyy)*dy*dz
                
                vv[i]      -= 6*( dy*mrrz - dz*mrry )*idr7
                vv[i+Nt]   -= 6*( dz*mrrx - dx*mrrz )*idr7
                vv[i+2*Nt] -= 6*( dx*mrry - dy*mrrx )*idr7
        return




"""Some utilities codes which do not fit anywhere else,
but are essential in simulations of active colloids,
and plotting flow and phoretic fields
"""

import  numpy as np
cimport numpy as np
import matplotlib.pyplot as plt
cimport cython
from libc.math cimport sqrt, pow, log
from cython.parallel import prange


cpdef irreducibleTensors(l, p, Y0=?)



cimport cython
from libc.math cimport sqrt, exp, pow, erfc, sin, cos
from cython.parallel import prange
import numpy as np
cimport numpy as np
cdef double PI = 3.14159265359

@cython.wraparound(False)
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
cdef class Rbm:

    cdef readonly int Nx, Ny, Nz, Np
    cdef readonly np.ndarray Mobility
    cdef readonly double Lx, Ly, Lz, a, facx, facy, facz, eta, mu, muv, mur


    cpdef mobilityTT(self, double [:] v, double [:] r, double [:] F)


    cpdef mobilityTR(   self, double [:] v, double [:] r, double [:] T)


    cpdef propulsionT2s(self, double [:] v, double [:] r, double [:] S)


    cpdef propulsionT3t(self, double [:] v, double [:] r, double [:] D)



    cpdef mobilityRT(self, double [:] o, double [:] r, double [:] F)


    cpdef mobilityRR(self, double [:] o, double [:] r, double [:] T)
    
    cpdef propulsionR2s(self, double [:] o, double [:] r, double [:] S)


    cpdef propulsionR3a(self, double [:] o, double [:] r, double [:] M)


    cpdef propulsionR4a(self, double [:] o, double [:] r, double [:] M)


    cpdef noiseTT(self, double [:] v, double [:] r)


    cpdef noiseRR(self, double [:] o, double [:] r)


## Flow at given points
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
@cython.wraparound(False)
cdef class Flow:
    cdef readonly double a, eta
    cdef readonly int Nt, Np


    cpdef flowField1s(self, double [:] vv, double [:] rt, double [:] r, double [:] F)

    cpdef flowField2a(  self, double [:] vv, double [:] rt, double [:] r, double [:] T)

    cpdef flowField2s(self, double [:] vv, double [:] rt, double [:] r, double [:] S)

    cpdef flowField3t(self, double [:] vv, double [:] rt, double [:] r, double [:] D)
    



"""Some utilities codes which do not fit anywhere else,
but are essential in simulations of active colloids,
and plotting flow and phoretic fields
"""

import  numpy as np
cimport numpy as np
cimport cython
from libc.math cimport sqrt, pow, log
from cython.parallel import prange
cdef double PI = 3.1415926535

DTYPE   = np.float
DTYP1   = np.int32
ctypedef np.float_t DTYPE_t 


def gridXY(dim, L, Ng):
    """
    Returns the grid in XY direction centered around zero
    ...

    Parameters
    ----------
    L  : float  
        Length of the grid 
    Ng : int 
         Number of grid points

    Examples
    --------
    An example of creating grid 
    
    >>> import numpy as np, pystokes 
    >>> dim, L, Ng = 3, 10, 32
    >>>  rr, vv = pystokes.utils.gridXY(dim, L, Ng)
    """

    Nt = Ng*Ng
    rr, vv = np.zeros(dim*Nt), np.zeros(dim*Nt)
    X, Y = np.meshgrid(np.linspace(-L, L, Ng), np.linspace(-L, L, Ng))
    rr[0:2*Nt] = np.concatenate((X.reshape(Ng*Ng), Y.reshape(Ng*Ng)))
    return rr, vv


def gridYZ(dim, L, Ng):
    """
    Returns the grid in YZ direction centered around zero
    ...

    Parameters
    ----------
    L  : float  
        Length of the grid 
    Ng : int 
         Number of grid points

    Examples
    --------
    An example of creating grid 
    
    >>> import numpy as np, pystokes 
    >>> dim, L, Ng = 3, 10, 32
    >>>  rr, vv = pystokes.utils.gridYZ(dim, L, Ng)
    """

    Nt = Ng*Ng
    rr, vv = np.zeros(dim*Nt), np.zeros(dim*Nt)
    X, Y = np.meshgrid(np.linspace(-L, L, Ng), np.linspace(-0, L, Ng))
    rr[Nt:3*Nt] = np.concatenate((X.reshape(Ng*Ng), Y.reshape(Ng*Ng)))
    return rr, vv



@cython.wraparound(False)
@cython.nonecheck(False)
@cython.boundscheck(False)
cpdef irreducibleTensors(l, p, Y0=1):
    """
    Uniaxial paramterization of the tensorial harmonics (Yl) of order l
    ...

    Parameters
    ----------
    l  : int 
        Tensorial Harmonics of order l
    p  : np.rrray 
        An array of size 3
        Axis along which the mode is paramterized
    Y0 : float 
        Strength of the mode
    
    returns: Yl - tensorialHarmonics of rank l
    """
    cdef int i, Np = int (np.size(p)/3)
    cdef double S0 = Y0
    YY = np.zeros((2*l+1)*Np, dtype=DTYPE)
    
    cdef double [:] p1 = p
    cdef double [:] Y1 = YY
    
    
    if l==0:
            YY = Y0
    
    if l==1:
            YY = Y0*p
    
    if l==2:
        for i in prange(Np, nogil=True):
            Y1[i + 0*Np] = S0*(p1[i]*p1[i]                   -(1.0/3))
            Y1[i + 1*Np] = S0*(p1[i + Np]*p1[i + Np] -(1.0/3))
            Y1[i + 2*Np] = S0*(p1[i]*p1[i + Np])
            Y1[i + 3*Np] = S0*(p1[i]*p1[i + 2*Np])
            Y1[i + 4*Np] = S0*(p1[i + Np]*p1[i + 2*Np])
    
    if l==3:
        for i in range(Np):
            YY[i]      = Y0*(p[i]*p[i]*p[i]                    - 3/5*p[i]);
            YY[i+Np]   = Y0*(p[i+Np]*p[i+Np]*p[i+Np]   - 3/5*p[i+Np]);
            YY[i+2*Np] = Y0*(p[i]*p[i]*p[i+Np]                 - 1/5*p[i+Np]);
            YY[i+3*Np] = Y0*(p[i]*p[i]*p[i+2*Np]       - 1/5*p[i+2*Np]);
            YY[i+4*Np] = Y0*(p[i]*p[i+Np]*p[1+Np]           -1/5* p[i]);
            YY[i+5*Np] = Y0*(p[i+Np]*p[i+Np]*p[i+2*Np]);
            YY[i+6*Np] = Y0*(p[i+Np]*p[i+Np]*p[i+2*Np] -1/5*p[i+2*Np]);
    return YY




def simulate(rp0, Tf, Npts, rhs, integrator='odeint', filename='this.mat',
            Ti=0, maxNumSteps=100000, **kwargs):
    """
    Simulates using choice of integrator
    
    ...

    Parameters
    ----------
    rp0 : np.array 
        Initial condition 
    Tf  : int 
         Final time 
    Npts: int 
        Number of points to return data 
    rhs : Python Function
        Right hand side to integrate 
    integrator: string 
        Default is 'odeint' of scipy. 
        Please set integrator='odeint' to use 
        the scipy.integrate's odeint (Default)
        Use integrator='solve_ivp' to use ivp  
        Use integrator='odespy-vode' to use vode
        from odespy (github.com/rajeshrinet/odespy).
        Use integrator='odespy-rkf45' to use RKF45  
        from odespy (github.com/rajeshrinet/odespy).
        Use integrator='odespy-rk4' to use RK4 from 
        odespy (github.com/rajeshrinet/odespy).     
        Alternatively, write your own integrator to 
        evolve the system in time and store the data.
    filename: string 
        filename to write the data. 
        Deafult is 'this.mat'
    """

    def dxdtEval(rp, t): 
        """
        returns the right hand side
        """
        return rhs(rp)
   

    if integrator=='odeint':
        from scipy.integrate import odeint
        time_points=np.linspace(Ti, Tf, Npts+1);
        X = odeint(dxdtEval, rp0, time_points, mxstep=maxNumSteps, **kwargs)

    elif integrator=='solve_ivp':
        from scipy.integrate import solve_ivp
        time_points=np.linspace(Ti, Tf, Npts+1)                                                          
        X = solve_ivp(lambda t, xt: dxdtEval(xt,t), [0,Tf], rp0, 
                         t_eval=time_points, **kwargs).y.T

    elif integrator=='odespy-vode':
        import odespy
        time_points=np.linspace(Ti, Tf, Npts+1);
        solver = odespy.Vode(dxdtEval, method = 'bdf', 
        atol=1E-7, rtol=1E-6, order=5, nsteps=maxNumSteps)
        solver.set_initial_condition(rp0)
        X, t = solver.solve(time_points, **kwargs) 

    elif integrator=='odespy-rkf45':
        import odespy
        time_points=np.linspace(Ti, Tf, Npts+1);
        solver = odespy.RKF45(dxdtEval)
        solver.set_initial_condition(rp0)
        X, t = solver.solve(time_points, **kwargs) 

    elif integrator=='odespy-rk4':
        import odespy
        time_points=np.linspace(Ti, Tf, Npts+1);
        solver = odespy.RK4(dxdtEval)
        solver.set_initial_condition(rp0)
        X, t = solver.solve(time_points, **kwargs) 

    else:
        raise Exception("Error: Integration method not found! \n \
                        Please set integrator='odeint' to use \n \
                        the scipy.integrate's odeint (Default)\n \
                        Use integrator='solve_ivp' to use ivp  \
                        Use integrator='odespy-vode' to use vode \
                        from odespy (github.com/rajeshrinet/odespy).\n \
                        Use integrator='odespy-rkf45' to use RKF45  \
                        from odespy (github.com/rajeshrinet/odespy).\n \
                        Use integrator='odespy-rk4' to use RK4 from \
                        odespy (github.com/rajeshrinet/odespy).     \n \
                        Alternatively, write your own integrator to \
                        evolve the system in time and store the data.\n")

    from scipy.io import savemat
    savemat(filename, {'X':X, 't':time_points})
    return




def initialCondition(Np, h0=3.1):
    '''
    Assigns initial condition. 
    To DO: Need to include options to assign possibilites
    '''
    rp0 = np.zeros(3*Np) 
    radius = np.sqrt(np.arange(Np)/float(Np))
    golden_angle = np.pi * (3 - np.sqrt(5))
    theta = golden_angle * np.arange(Np)
    
    points = np.zeros((Np, 2)) 
    points[:,0] = np.cos(theta)
    points[:,1] = np.sin(theta)
    points *= radius.reshape((Np, 1)) 
    points = points*2.1*np.sqrt(Np)
    
    points[:,0] += 2*(2*np.random.random(Np)-np.ones(Np)) 
    points[:,1] += 2*(2*np.random.random(Np)-np.ones(Np)) 
    
    
    rp0[0:Np]          = 1.5*points[:, 0]
    rp0[Np:2*Np]   = 1.5*points[:, 1]
    rp0[2*Np:3*Np] = h0
    ##rp0[0:3*Np] = ar*(2*np.random.random(3*Np)-np.ones(3*Np)) + rp0[0:3*Np]
    #
    #rp0[3*Np:4*Np] = np.zeros(Np)
    #rp0[4*Np:5*Np] = np.zeros(Np)
    #rp0[5*Np:6*Np] = -np.ones(Np)
    return rp0 




def plotLogo():
    """Plots the PyStokes Logo"""
    import pystokes 
    import numpy as np, matplotlib.pyplot as plt
    # particle radius, fluid viscosity, and number of particles
    b, eta, Np = 1.0, 1.0/6.0, 4
    
    #initialise
    r = np.zeros(3*Np);  r[2*Np:3*Np] = 2
    p = np.zeros(3*Np);  p[2*Np:3*Np] = 1
    
    r[2*Np]=2.8;  r[Np:2*Np]=np.linspace(-3.2, 3.2, Np);
    # irreducible coeffcients
    F1s = pystokes.utils.irreducibleTensors(1, p)
    V2s = pystokes.utils.irreducibleTensors(2, p)
    V3t = pystokes.utils.irreducibleTensors(1, p) 
    
    # space dimension , extent , discretization
    dim, L, Ng = 3, 8, 100
    
    #Instantiate the Flow class
    wFlow = pystokes.unbounded.Flow(radius=b, particles=Np, viscosity=eta, gridpoints=Ng*Ng)
    wFlow = pystokes.wallBounded.Flow(radius=b, particles=Np, viscosity=eta, gridpoints=Ng*Ng)
    #iFlow = pystokes.interface.Flow(radius=b, particles=Np, viscosity=eta, gridpoints=Ng*Ng)
    
    # create the grid
    rr, vv = pystokes.utils.gridYZ(dim, L, Ng); 
    wFlow.flowField1s(vv, rr, r, F1s)  
    
    plt.figure(figsize=(26, 8));  Nt=Ng*Ng
    
    yy, zz = rr[Nt:2*Nt].reshape(Ng, Ng), rr[2*Nt:3*Nt].reshape(Ng, Ng)
    vy, vz = vv[Nt:2*Nt].reshape(Ng, Ng), vv[2*Nt:3*Nt].reshape(Ng, Ng)
    density=0.75; arrowSize=4; mask=0.6; ms=36; offset=1e-6
    plt.streamplot(yy, zz, vy, vz, color="gray", arrowsize =arrowSize, density=density, linewidth=4.4)
    
    for i in range(Np):
            plt.plot(r[i+Np], r[i+2*Np], 'o', mfc='snow', mec='darkslategray', ms=4.8*ms, mew=5.2)
    plt.grid()
    
    ww=0.3
    plt.ylim(-ww, np.max(zz))
    plt.axhspan(-ww, ww, facecolor='gray');
    plt.axis('off')
    plt.text(r[Np]-.5, r[2*Np]-.5, 'Py', fontsize=100);
    plt.text(r[Np+1]-.46, r[2*Np+1]-.54, 'St', fontsize=111);
    plt.text(r[Np+2]-.58, r[2*Np+2]-.54, 'ok', fontsize=111);
    plt.text(r[Np+3]-.46, r[2*Np+3]-.54, 'es', fontsize=111);




def plotStreamlinesXY(vv, rr, r, density=0.82, arrowSize=1.2, mask=0.6, ms=36, offset=1e-6, title='None'):
    """
    Plots streamlines in XY plane given the position and velocity; 
    vv: one dimensional arrays of velocity
    rr: one dimensional arrays of positions where velocity is computed
    """
    import matplotlib.pyplot as plt
    Np, Nt = int(np.size(r)/3), int(np.size(rr)/3);  Ng=int(np.sqrt(Nt))
    xx, yy = rr[0:Nt].reshape(Ng, Ng), rr[Nt:2*Nt].reshape(Ng, Ng)
    vx, vy = vv[0:Nt].reshape(Ng, Ng), vv[Nt:2*Nt].reshape(Ng, Ng)

    for i in range(Np):
        plt.plot(r[i], r[i+Np], 'o', mfc='snow', mec='darkslategray', ms=ms, mew=4 )   

    spd = np.hypot(vx, vy)
    rr  = np.hypot(xx-r[0], yy-r[1])
    spd[rr<mask]=0;  spd+=offset
    plt.pcolormesh(xx, yy, np.log(spd), cmap=plt.cm.gray_r, shading='interp')

    st=11;  sn=st*st;  a0=2;  ss=np.zeros((2, st*st))
    for i in range(st):
        for j in range(st):
            ii = i*st+j
            ss[0, ii]      = a0*(-5 + i)
            ss[1, ii]      = a0*(-5 + j)

    plt.streamplot(xx, yy, vx, vy, color="black", arrowsize =arrowSize, arrowstyle='->', start_points=ss.T, density=density)
    plt.xlim(np.min(xx), np.max(xx))
    plt.ylim(np.min(yy), np.max(yy))
    plt.axis('off')
    
    if title==str('None'):
        pass
    elif title==str('1s'):
        plt.title('$l\sigma=1s$', fontsize=26);
    elif title==str('2s'):
        plt.title('$l\sigma=2s$', fontsize=26);
    elif title==str('3t'):
        plt.title('$l\sigma=3t$', fontsize=26);
    else:
        plt.title(title, fontsize=26);




def plotStreamlinesYZ(vv, rr, r, density=0.795, arrowSize=1.2, mask=0.6, ms=36, offset=1e-6, title='None'):
    """
    Plots streamlines in YZ plane given the position and velocity; 
    vv: one dimensional arrays of velocity
    rr: one dimensional arrays of positions where velocity is computed
    """
    import matplotlib.pyplot as plt
    Np, Nt = int(np.size(r)/3), int(np.size(rr)/3);  Ng=int(np.sqrt(Nt))
    yy, zz = rr[Nt:2*Nt].reshape(Ng, Ng), rr[2*Nt:3*Nt].reshape(Ng, Ng)
    vy, vz = vv[Nt:2*Nt].reshape(Ng, Ng), vv[2*Nt:3*Nt].reshape(Ng, Ng)

    for i in range(Np):
        plt.plot(r[i+Np], r[i+2*Np], 'o', mfc='snow', mec='darkslategray', ms=ms, mew=4 )   

    spd = np.hypot(vy, vz)
    rr  = np.hypot(yy-r[1], zz-r[2])
    spd[rr<mask]=0;  spd+=offset
    plt.pcolormesh(yy, zz, np.log(spd), cmap=plt.cm.gray_r, shading='interp')

    st=11;  sn=st*st;  a0=2;  ss=np.zeros((2, st*st))
    for i in range(st):
        for j in range(st):
            ii = i*st+j
            ss[0, ii]      = a0*(-5 + i)
            ss[1, ii]      = 1*(0 + j)
    plt.streamplot(yy, zz, vy, vz, color="black", arrowsize =arrowSize, arrowstyle='->', start_points=ss.T, density=density)
    plt.xlim(np.min(yy), np.max(yy))
    plt.ylim(np.min(zz), np.max(zz))
    plt.axis('off')
    
    if title==str('None'):
        pass
    elif title==str('1s'):
        plt.title('$l\sigma=1s$', fontsize=26);
    elif title==str('2s'):
        plt.title('$l\sigma=2s$', fontsize=26);
    elif title==str('3t'):
        plt.title('$l\sigma=3t$', fontsize=26);
    else:
        plt.title(title, fontsize=26);




def plotStreamlinesYZsurf(vv, rr, r, density=0.8, arrowSize=1.2, mask=0.6, ms=36, offset=1e-6, title='None'):
    """
    Plots streamlines in YZ plane given the position and velocity; The surface is also plotted
    vv: one dimensional arrays of velocity
    rr: one dimensional arrays of positions where velocity is computed
    """
    import matplotlib.pyplot as plt
    Np, Nt = int(np.size(r)/3), int(np.size(rr)/3);  Ng=int(np.sqrt(Nt))
    yy, zz = rr[Nt:2*Nt].reshape(Ng, Ng), rr[2*Nt:3*Nt].reshape(Ng, Ng)
    vy, vz = vv[Nt:2*Nt].reshape(Ng, Ng), vv[2*Nt:3*Nt].reshape(Ng, Ng)

    for i in range(Np):
        plt.plot(r[i+Np], r[i+2*Np], 'o', mfc='snow', mec='darkslategray', ms=ms, mew=4 )   
    
    spd = np.hypot(vy, vz)
    rr  = np.hypot(yy-r[1], zz-r[2])
    spd[rr<mask]=0;  spd+=offset
    plt.pcolormesh(yy, zz, np.log(spd), cmap=plt.cm.gray_r, shading='interp')
    
    st=11;  sn=st*st;  a0=2;  ss=np.zeros((2, st*st))
    for i in range(st):
        for j in range(st):
            ii = i*st+j
            ss[0, ii]      = a0*(-5 + i)
            ss[1, ii]      = 1*(0 + j)
    plt.streamplot(yy, zz, vy, vz, color="black", arrowsize =arrowSize, arrowstyle='->', start_points=ss.T, density=density)
    plt.xlim(np.min(yy), np.max(yy))
    ww=0.3
    plt.ylim(-ww, np.max(zz))
    plt.axhspan(-ww, ww, facecolor='black');
    plt.axis('off')
    
    if title==str('None'):
        pass
    elif title==str('1s'):
        plt.title('$l\sigma=1s$', fontsize=26);
    elif title==str('2s'):
        plt.title('$l\sigma=2s$', fontsize=26);
    elif title==str('3t'):
        plt.title('$l\sigma=3t$', fontsize=26);
    else:
        plt.title(title, fontsize=26);





def plotContoursYZ(vv, rr, r, density=1.2, arrowSize=1.2, mask=0.6, ms=36, offset=1e-6, title='None'):
    """
    Plots streamlines in YZ plane given the position and velocity; 
    vv: one dimensional arrays of velocity
    rr: one dimensional arrays of positions where velocity is computed
    """
    import matplotlib.pyplot as plt
    Np, Nt = int(np.size(r)/3), int(np.size(rr)/3);  Ng=int(np.sqrt(Nt))
    yy, zz = rr[Nt:2*Nt].reshape(Ng, Ng), rr[2*Nt:3*Nt].reshape(Ng, Ng)

    for i in range(Np):
        plt.plot(r[i+Np], r[i+2*Np], 'o', mfc='snow', mec='darkslategray', ms=ms, alpha=0.8, mew=4 )   

    spd = np.sqrt(vv[0:Nt]*vv[0:Nt]).reshape(Ng, Ng)
    spd+=offset
    cp=plt.pcolor(yy, zz, np.log(spd), cmap=plt.cm.gray_r)
    cp=plt.contour(yy, zz, np.log(spd), colors='#778899', linestyles='solid')

    plt.xlim(np.min(yy), np.max(yy))
    ww=0.
    plt.ylim(-ww, np.max(zz))
    plt.axis('off')
    if title==str('None'):
        pass 
    elif title==str('m=0'):
        plt.title('$m=0$', fontsize=26);
        #plt.ylabel('Unbounded domain', fontsize=20)
    elif title==str('m=1'):
        plt.title('$m=1$', fontsize=26);
    elif title==str('m=2'):
        plt.title('$m=2$', fontsize=26);
    else:
        plt.title(title, fontsize=26);




def plotContoursYZsurf(vv, rr, r, density=1.2, arrowSize=1.2, mask=0.6, ms=36, offset=1e-6, title='None'):
    """
    Plots streamlines in YZ plane given the position and velocity; 

    vv: one dimensional arrays of velocity
    rr: one dimensional arrays of positions where velocity is computed
    """
    import matplotlib.pyplot as plt
    Np, Nt = int(np.size(r)/3), int(np.size(rr)/3);  Ng=int(np.sqrt(Nt))
    yy, zz = rr[Nt:2*Nt].reshape(Ng, Ng), rr[2*Nt:3*Nt].reshape(Ng, Ng)

    for i in range(Np):
        plt.plot(r[i+Np], r[i+2*Np], 'o', mfc='snow', mec='darkslategray', ms=ms, alpha=0.8, mew=4 )   

    spd = np.sqrt(vv[0:Nt]*vv[0:Nt]).reshape(Ng, Ng)
    spd+=offset
    cp=plt.pcolor(yy, zz, np.log(spd), cmap=plt.cm.gray_r)
    cp=plt.contour(yy, zz, np.log(spd), colors='#778899', linestyles='solid')

    plt.xlim(np.min(yy), np.max(yy))
    ww=0.3
    plt.ylim(-ww, np.max(zz))
    plt.axhspan(-ww, ww, facecolor='black');
    plt.axis('off')
    
    if title==str('None'):
        pass 
    elif title==str('l=0'):
        plt.title('$l=0$', fontsize=26);
        plt.ylabel('Plane no-flux wall', fontsize=20)
    elif title==str('l=1'):
        plt.title('$l=1$', fontsize=26);
    elif title==str('l=0 and l=1'):
        plt.title('$l=0$ and $l=1$', fontsize=26);
    else:
        plt.title(title, fontsize=26);
    #elif title==str('0'):
    #    plt.title('title, fontsize=26);
    #elif title==str('1'):
    #    plt.title('$l=1$', fontsize=26);



def plotTrajectory(twoBodyDynamics, T, bins=50):
    import matplotlib.pyplot as plt
    f=plt.figure(figsize=(24, 11), edgecolor='gray', linewidth=4);  
    ax=f.add_subplot(221);
    x1, x2 = twoBodyDynamics(T[0])
    plt.hist(np.abs(x1-x2), bins=bins, density=True, color='gray', alpha=0.5)
    plt.title("Temperature  = %s"%T[0], fontsize=24)
    plt.ylabel('Separation of colloids', fontsize=28)
    plt.xticks(fontsize=24); plt.yticks(fontsize=24)
    plt.yticks([])
    
    ax=f.add_subplot(223);
    plt.plot(x1, '--', label="colloid 1", color="dimgray", lw=3.2)
    plt.plot(x2, '-', label="colloid 2", color="slategray", lw=3.2)
    plt.legend(fontsize=24, loc='upper right'); 
    plt.xticks(fontsize=24); 
    plt.ylabel('Position of colloids', fontsize=28)
    plt.xticks([])
    ax.yaxis.tick_right()
    plt.yticks(fontsize=24)
    
    ax=f.add_subplot(222);
    x1, x2 = twoBodyDynamics(T[1]);  data=np.abs(x1-x2)
    #plt.hist(np.abs(x1-x2), 64, density=True, color='gray')
    y,binEdges=np.histogram(data,bins=bins)
    bincenters = 0.5*(binEdges[1:]+binEdges[:-1])
    plt.hist(data, bins=bins, label='simulation', color='gray', alpha=0.5);
    plt.gca().set_xlim(left=2.0)
   
    aa=0 
    bincenters = np.sort(bincenters)
    for i in range(10):
        if bincenters[i]<1.9:
            aa+=1

    xx = bincenters[aa:];   rr=(xx**2 + 4*2.5*2.5)**(-1.5)
    yy = np.log(y[aa:]);    coefficients = np.polyfit(rr, yy, 1)
    polynomial = np.poly1d(coefficients); ys = polynomial(rr)
    plt.plot(xx[0:], np.exp(ys)[0:], '-', label='analytical', lw=3.2, color='dimgray')
    plt.title("Temperature  = %s"%T[1], fontsize=24)
    plt.xticks(fontsize=24); plt.yticks(fontsize=24)
    plt.yticks([]); plt.legend(fontsize=22)
    
    ax=f.add_subplot(224);
    plt.plot(x1, '--', label="colloid 1", color="dimgray", lw=2)
    plt.plot(x2, '-', label="colloid 2", color="slategray", lw=2)
    plt.legend(fontsize=24); plt.xticks(fontsize=24); plt.yticks(fontsize=24)
    plt.xticks([])
    ax.yaxis.tick_right()
    plt.yticks(fontsize=24)
    


def createCircle(R, alpha=1):
    import matplotlib.pyplot as plt
    return plt.Circle((0,0), radius= R, color='silver', alpha=alpha) 



def showShape(patch):
    import matplotlib.pyplot as plt
    ax=plt.gca()
    ax.add_patch(patch)
    plt.axis('scaled')
    #plt.show()



def plotConfigs(t=[0,1], ms=36, tau=1, filename='None'):
    import matplotlib.pyplot as plt
    tNew= np.asanyarray(tau)*t
    from scipy.io import loadmat 
    data  = loadmat(filename)
    X     = data['X']
    xx = int(np.size(X[0,:]));  Np = int(xx/6);   
    color='gray'


    plt.figure(figsize=(28, 10), edgecolor='gray', linewidth=4)

    ll = np.abs(np.max(X[t[0], 0:Np]))+5
    if np.size(t)==1:
        c = createCircle(ll, alpha=1);     showShape(c)
        plt.scatter(X[t[0], 0:Np], X[t[0], Np:2*Np], s=ms, c=color, edgecolors='darkslategray', mew=4); 
        plt.xlim(-ll, ll); plt.ylim(-ll, ll); plt.axis('off');
    
    elif np.size(t)==2:
        plt.subplot(141); 
        c = createCircle(ll, alpha=1);     showShape(c)
        plt.scatter(X[t[0], 0:Np], X[t[0], Np:2*Np], s=ms, c=color, edgecolors='darkslategray'); plt.axis('off');
        plt.subplot(142);                                                                     
        c = createCircle(ll, alpha=1);     showShape(c)
        plt.scatter(X[t[1], 0:Np], X[t[1], Np:2*Np], s=ms, c=color, edgecolors='darkslategray'); plt.axis('off');
                                                                                              
    elif np.size(t)==3:                                                                       
        plt.subplot(141);                                                                     
        c = createCircle(ll, alpha=1);     showShape(c)
        plt.scatter(X[t[0], 0:Np], X[t[0], Np:2*Np], s=ms, c=color, edgecolors='darkslategray'); plt.axis('off');
                                                                                              
        plt.subplot(142);                                                                     
        c = createCircle(ll, alpha=1);     showShape(c)
        plt.scatter(X[t[1], 0:Np], X[t[1], Np:2*Np], s=ms, c=color, edgecolors='darkslategray'); plt.axis('off');
                                                                                              
        plt.subplot(143);                                                                     
        c = createCircle(ll, alpha=1);     showShape(c)
        plt.scatter(X[t[2], 0:Np], X[t[2], Np:2*Np], s=ms, c=color, edgecolors='darkslategray'); plt.axis('off');

    elif np.size(t)==4:
        plt.subplot(141); 
        c = createCircle(ll, alpha=.1);     showShape(c)
        plt.scatter(X[t[0], 0:Np], X[t[0], Np:2*Np], s=ms, c=color, edgecolors='darkslategray'); 
        plt.xlim(-ll, ll); plt.ylim(-ll, ll); plt.axis('off'); plt.title(r'Time=%d$\tau$'%tNew[0], fontsize=32)
        
        plt.subplot(142); 
        c = createCircle(ll, alpha=.1);     showShape(c)
        plt.scatter(X[t[1], 0:Np], X[t[1], Np:2*Np], s=ms, c=color, edgecolors='darkslategray'); 
        plt.xlim(-ll, ll); plt.ylim(-ll, ll); plt.axis('off'); plt.title(r'Time=%d$\tau$'%tNew[1], fontsize=32)
        
        plt.subplot(143); 
        c = createCircle(ll, alpha=.1);     showShape(c)
        plt.scatter(X[t[2], 0:Np], X[t[2], Np:2*Np], s=ms, c=color, edgecolors='darkslategray');
        plt.xlim(-ll, ll); plt.ylim(-ll, ll); plt.axis('off'); plt.title(r'Time=%d$\tau$'%tNew[2], fontsize=32)
        
        plt.subplot(144); 
        c = createCircle(ll, alpha=.1);     showShape(c)
        plt.scatter(X[t[3], 0:Np], X[t[3], Np:2*Np], s=ms, c=color, edgecolors='darkslategray');
        plt.xlim(-ll, ll); plt.ylim(-ll, ll); plt.axis('off'); plt.title(r'Time=%d$\tau$'%tNew[3], fontsize=32)

        
    elif np.size(t)==5:
        plt.subplot(151); 
        c = createCircle(ll, alpha=1);     showShape(c)
        plt.scatter(X[t[0], 0:Np], X[t[0], Np:2*Np], s=ms, c=color, edgecolors='darkslategray'); 
        plt.xlim(-ll, ll); plt.ylim(-ll, ll); plt.axis('off'); plt.title('Time=%d$\tau$'%tNew[0], fontsize=26)
        
        plt.subplot(152); 
        c = createCircle(ll, alpha=1);     showShape(c)
        plt.scatter(X[t[1], 0:Np], X[t[1], Np:2*Np], s=ms, c=color, edgecolors='darkslategray'); 
        plt.xlim(-ll, ll); plt.ylim(-ll, ll); plt.axis('off'); plt.title('Time=%d$\tau$'%tNew[1], fontsize=26)
        
        plt.subplot(153); 
        c = createCircle(ll, alpha=1);     showShape(c)
        plt.scatter(X[t[2], 0:Np], X[t[2], Np:2*Np], s=ms, c=color, edgecolors='darkslategray');
        plt.xlim(-ll, ll); plt.ylim(-ll, ll); plt.axis('off'); plt.title('Time=%d$\tau$'%tNew[2], fontsize=26)
        
        plt.subplot(154); 
        c = createCircle(ll, alpha=1);     showShape(c)
        plt.scatter(X[t[3], 0:Np], X[t[3], Np:2*Np], s=ms, c=color, edgecolors='darkslategray');
        plt.xlim(-ll, ll); plt.ylim(-ll, ll); plt.axis('off'); plt.title('Time=%d$\tau$'%tNew[3], fontsize=26)
        
        plt.subplot(155); 
        c = createCircle(ll, alpha=1);     showShape(c)
        plt.scatter(X[t[4], 0:Np], X[t[4], Np:2*Np], s=ms, c=color, edgecolors='darkslategray');
        plt.xlim(-ll, ll); plt.ylim(-ll, ll); plt.axis('off'); plt.title('Time=%d$\tau$'%tNew[4], fontsize=26)
    return 



@cython.wraparound(False)
@cython.nonecheck(False)
@cython.boundscheck(False)
cpdef couplingTensors(l, p, M0=1):
    """
    Uniaxial paramterization of the tensorial harmonics (Yl) of order l
    l  : tensorialHarmonics of order l
    p  : axis along which the mode is paramterized
    M0 : strength of the mode

    returns: Yl - tensorial harmonics of rank l
    """
    cdef int i, Np = int (np.size(p)/3)
    cdef double S0 = M0
    MM = np.zeros((2*l+1)*Np, dtype=DTYPE)

    cdef double [:] p1 = p
    cdef double [:] Y1 = MM


    if l==0:
        MM = -M0

    if l==1:
        MM = -M0*p
    
    if l==2:
        for i in prange(Np, nogil=True):
            Y1[i + 0*Np] = S0*(p1[i]*p1[i]           -(1.0/3))
            Y1[i + 1*Np] = S0*(p1[i + Np]*p1[i + Np] -(1.0/3))
            Y1[i + 2*Np] = S0*(p1[i]*p1[i + Np])
            Y1[i + 3*Np] = S0*(p1[i]*p1[i + 2*Np])
            Y1[i + 4*Np] = S0*(p1[i + Np]*p1[i + 2*Np])

    if l==3:
        for i in range(Np):
            MM[i]      = M0*(p[i]*p[i]*p[i]            - 3/5*p[i]);
            MM[i+Np]   = M0*(p[i+Np]*p[i+Np]*p[i+Np]   - 3/5*p[i+Np]);
            MM[i+2*Np] = M0*(p[i]*p[i]*p[i+Np]         - 1/5*p[i+Np]);
            MM[i+3*Np] = M0*(p[i]*p[i]*p[i+2*Np]       - 1/5*p[i+2*Np]);
            MM[i+4*Np] = M0*(p[i]*p[i+Np]*p[1+Np]       -1/5* p[i]);
            MM[i+5*Np] = M0*(p[i+Np]*p[i+Np]*p[i+2*Np]);
            MM[i+6*Np] = M0*(p[i+Np]*p[i+Np]*p[i+2*Np] -1/5*p[i+2*Np]);
    return MM




def plotPhoreticField(l, c0=1):
    """
    l   : mode of phoretic field
    vls0: strength of the mode
    """

    theta = np.linspace(0, np.pi, 128)
    phi = np.linspace(0, 2*np.pi, 128)
    theta, phi = np.meshgrid(theta, phi)

    x = np.sin(theta) * np.cos(phi)
    y = np.sin(theta) * np.sin(phi)
    z = np.cos(theta)
    
    speedVls = z
    
    # Set the aspect ratio to 1 so our sphere looks spherical
    import matplotlib.pyplot as plt
    fig = plt.figure(figsize=(4, 4))
    ax = fig.add_subplot(111, projection='3d')

    if l==0:
        speedVls = c0 + z*0
        ax.plot_surface(x, y, z,  rstride=1, cstride=1, facecolors=plt.cm.RdBu_r(speedVls))
        ax.set_axis_off()
        plt.show()
    
    elif l==1:
        speedVls = c0*np.cos(theta)
        ax.plot_surface(x, y, z,  rstride=1, cstride=1, facecolors=plt.cm.RdBu_r(speedVls))
        ax.set_axis_off()
        plt.show()
    
    elif l==2:
        speedVls = 4*c0*(np.cos(theta)*np.cos(theta) - 1.0/3)
        ax.plot_surface(x, y, z,  rstride=1, cstride=1, facecolors=plt.cm.RdBu_r(speedVls))
        ax.set_axis_off()
        plt.show()
    
    elif l==3:
        speedVls = 4*c0*(np.cos(theta)*np.cos(theta)*np.cos(theta) - 3*np.cos(theta)/5.0)
        ax.plot_surface(x, y, z,  rstride=1, cstride=1, facecolors=plt.cm.RdBu_r(speedVls))
        ax.set_axis_off()
        plt.show()

    else:
        print('Not yet implemented...')



cpdef MSD2(x1, y1, int tdev):
    """
    Mean square displacement in two dimensions 
    x1  : time series of x-coordinate
    y1  : time series of y-coordinate
    tdev: time difference

    returns: dispV: a vector of square of displacement
    """
    cdef int i, j, Nt
    cdef double dr2, dx, dy
    dispV = np.zeros(tdev)
    
    cdef double [:] xt = x1
    cdef double [:] yt = y1
    cdef double [:] d2 = dispV

    Nt = x1.size 

    for i in range(tdev): 
        dr2=0
        for j in range(Nt-i-1):
            dx = xt[i+j+1] - xt[j]
            dy = yt[i+j+1] - yt[j]
            dr2 += dx*dx + dy*dy 
        d2[i] = dr2/(Nt-i-1)
    return dispV



cimport cython
from libc.math cimport sqrt
from cython.parallel import prange
import numpy as np
cimport numpy as np
cdef double PI = 3.14159265359



@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
@cython.wraparound(False)
cdef class Rbm:                         
    """
    Rigid body motion (RBM)
    
    ...

    ----------
    radius: float
        Radius of the particles (a).    
    particles: int
        Number of particles (Np)
    viscosity: float 
        Viscosity of the fluid (eta)
    """
 

    def __init__(self, a, Np, eta):
        self.a  = a                 # radius of the particles
        self.Np = Np                # number of particles
        self.eta = eta                # number of particles

    cpdef mobilityTT(self, double [:] v, double [:] r, double [:] F, double H):
        """
        Compute velocity due to body forces using :math:`v=\mu^{TT}\cdot F` 
        ...

        Parameters
        ----------
        v: np.array
            An array of velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        F: np.array
            An array of forces
            An array of size 3*Np,
        H: float 
            Height of the Hele-Shaw cell 
        """

        cdef int i, j, Np=self.Np, xx=2*Np
        cdef double dx, dy, dz, idr, idr2, Fdotidr2, h2, hsq, tempF
        cdef double vx, vy, vz, tH = 2*H
        cdef double mu = 1.0/(6*PI*self.eta*self.a), mu1 = mu*self.a*0.75, a2=self.a*self.a/3.0
        cdef double fac0 = 3/(PI*self.eta*H*H*H), fac1, fac2
 
        for i in prange(Np, nogil=True):
            vx=0; vy=0; vz=0;
            for j in range(Np):
                dx = r[i]    - r[j]
                dy = r[i+Np]  - r[j+Np]
                h2  =  2*r[j+xx]; hsq=r[j+xx]*r[j+xx]
                if i!=j:
                    idr = 1.0/sqrt( dx*dx + dy*dy )
                    idr2=idr*idr
                    Fdotidr2 = (F[j] * dx + F[j+Np] * dy )*idr2
                    #
                    fac1 = fac0*(H-r[i+xx])*(H-r[j+xx])*r[i+xx]
                    vx += fac1*(0.5*F[j]    + Fdotidr2*dx)*idr2 
                    vy += fac1*(0.5*F[j+Np] + Fdotidr2*dy)*idr2 
                    
            v[i  ]  += mu*F[i]    + mu1*vx 
            v[i+Np] += mu*F[i+Np] + mu1*vy
        return 
    
    
    cpdef propulsionT2s(self, double [:] v, double [:] r, double [:] S, double H):
        """
        Compute velocity due to 2s mode of the slip 
        ...

        Parameters
        ----------
        v: np.array
            An array of velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        S: np.array
            An array of forces
            An array of size 5*Np,
        H: float 
            Height of the Hele-Shaw cell 
        """

        cdef int Np=self.Np, i, j, xx=2*Np, xx1=3*Np , xx2=4*Np   
        cdef double dx, dy, dz, idr, idr2, idr4, idr6, idr7, aidr2, trS, h2, hsq
        cdef double sxx, syy, szz, sxy, syx, syz, szy, sxz, szx, srr, srx, sry, srz
        cdef double Sljrlx, Sljrly, Sljrlz, Sljrjx, Sljrjy, Sljrjz 
        cdef double vx, vy, vz, mus =-(28.0*self.a**3)/24, tH = 2*H
        cdef double fac0 = 3/(PI*self.eta*H*H*H), fac1, fac2

        for i in prange(Np, nogil=True):
            vx=0; vy=0;   vz=0;
            for j in  range(Np):
                h2 = 2*r[j+xx]; hsq = r[j+xx]*r[j+xx];
                sxx = S[j]  ; syy = S[j+Np]; szz = -sxx-syy;
                sxy = S[j+xx]; syx = sxy;
                sxz = S[j+xx1]; szx = sxz;
                syz = S[j+xx2]; szy = syz;
                dx = r[i]   - r[j]
                dy = r[i+Np] - r[j+Np]
                if i!=j:
                    idr  = 1.0/sqrt( dx*dx + dy*dy);
                    idr4 = idr*idr*idr*idr; idr6 = idr4*idr*idr; 
                    fac1 = fac0*(H-r[i+xx])*(H-r[j+xx])*r[i+xx]

                    srx = fac1*(sxx*dx +  sxy*dy )*idr4; 
                    sry = fac1*(sxy*dx +  syy*dy )*idr4;
                    srz = fac1*(sxz*dx +  syz*dy )*idr4;
                    srr = fac1*(sxx*dx*dx + syy*dy*dy + 2*sxy*dx*dy )*idr6;
                    
                    vx += -2*srx + 4*srr*dx;
                    vy += -2*sry + 4*srr*dy;
                     
            v[i]    += vx*mus
            v[i+Np] += vy*mus
            v[i+xx] += vz*mus
        return

   
    cpdef propulsionT3t(self, double [:] v, double [:] r, double [:] D, double H):
        """
        Compute velocity due to 3t mode of the slip 
        ...

        Parameters
        ----------
        v: np.array
            An array of velocities
            An array of size 3*Np,
        r: np.array
            An array of positions
            An array of size 3*Np,
        D: np.array
            An array of forces
            An array of size 3*Np,
        H: float 
            Height of the Hele-Shaw cell 
        """

        cdef int Np=self.Np, i, j, xx=2*Np
        cdef double dx, dy, dz, idr, idr2, idr5, Ddotidr, tempD, hsq, h2, tH=2*H
        cdef double vx, vy, vz, mud = 3.0*self.a*self.a*self.a/5
        cdef double fac0 = 6/(PI*self.eta*H*H*H), fac1, fac2

        for i in prange(Np, nogil=True):
            vx=0; vy=0; vz=0;
            for j in range(Np):
                dx = r[i]    - r[j]
                dy = r[i+Np]  - r[j+Np]
                h2  =  2*r[j+xx]
                if i!=j:
                    idr = 1.0/sqrt( dx*dx + dy*dy)
                    idr2=idr*idr
                    Ddotidr = (D[j]*dx + D[j+Np]*dy)*idr2
                    
                    fac1 = -fac0*r[j+xx]*(H-r[j+xx])
                    vx += fac1*(0.5*D[j]    - Ddotidr*dx)*idr2
                    vy += fac1*(0.5*D[j+Np] - Ddotidr*dy)*idr2
            v[i]    += vx*mud
            v[i+Np] += vy*mud
            v[i+xx] += vz*mud





@cython.wraparound(False)
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
cdef class Flow:
    """
    Flow at given points
    
    ...

    Parameters
    ----------
    radius: float
        Radius of the particles.    
    particles: int
        Number of particles 
    viscosity: viscosity of the fluid 
    gridpoints: int 
        Number of grid points

    """

    def __init__(self, radius=1, particles=1, viscosity=1, gridpoints=32):
        self.a  = radius
        self.Np = particles
        self.Nt = gridpoints
        self.eta= viscosity


    cpdef flowField1s(self, double [:] vv, double [:] rt, double [:] r, double [:] F, double H):
        """
        Compute flow field at field points due body forces
        ...

        Parameters
        ----------
        vv: np.array
            An array of flow at field points
            An array of size 3*Nt,
        rt: np.array
            An array of field points
            An array of size 3*Nt,
        r: np.array
            An array of positions
            An array of size 3*Np,
        F: np.array
            An array of body force
            An array of size 3*Np,
        H: float 
            Height of the Hele-Shaw cell 
        """

        cdef int i, j, Np=self.Np, xx=2*Np, Nt=self.Nt
        cdef double dx, dy, dz, idr, idr3, idr5, Fdotidr, h2, hsq, tempF
        cdef double vx, vy, vz, tH = 2*H
        cdef double mu = 1.0/(6*PI*self.eta*self.a), mu1 = mu*self.a*0.75, a2=self.a*self.a/3.0
        cdef double fac0 = 3/(PI*self.eta*H*H*H), fac1, fac2
 
        for i in prange(Nt, nogil=True):
            vx=0; vy=0; vz=0;
            for j in range(Np):
                dx = rt[i]    - r[j]
                dy = rt[i+Nt]  - r[j+Np]
                idr = 1.0/sqrt( dx*dx + dy*dy )
                Fdotidr = (F[j] * dx + F[j+Np] * dy )*idr*idr
                #
                fac1 = (H-rt[i+2*Nt])*(H-r[j+xx])*rt[i+2*Nt]*r[j+xx]
                vx += fac1*(0.5*F[j]    - Fdotidr*dx)*idr*idr 
                vy += fac1*(0.5*F[j+Np] - Fdotidr*dy)*idr*idr
                    
            vv[i  ]  += mu1*vx 
            vv[i+Nt] += mu1*vy 



    cpdef flowField2s(self, double [:] vv, double [:] rt, double [:] r, double [:] S, double H):
        """
        Compute flow field at field points due to 2s mode of th slip 
        ...

        Parameters
        ----------
        vv: np.array
            An array of flow at field points
            An array of size 3*Nt,
        rt: np.array
            An array of field points
            An array of size 3*Nt,
        r: np.array
            An array of positions
            An array of size 3*Np,
        S: np.array
            An array of 2s mode of the slip 
            An array of size 5*Np,
        H: float 
            Height of the Hele-Shaw cell 
        """
        cdef int i, j, Np=self.Np, xx=2*Np, Nt=self.Nt, xx1=3*Np, xx2=4*Np
        cdef double dx, dy, dz, idr, idr2, idr4, idr6, idr7, aidr2, trS, h2, hsq
        cdef double sxx, syy, szz, sxy, syx, syz, szy, sxz, szx, srr, srx, sry, srz
        cdef double Sljrlx, Sljrly, Sljrlz, Sljrjx, Sljrjy, Sljrjz 
        cdef double vx, vy, vz, mus =-(28.0*self.a**3)/24, tH = 2*H
        cdef double fac0 = 3/(PI*self.eta*H*H*H), fac1, fac2

 
        for i in prange(Nt, nogil=True):
            vx=0; vy=0; vz=0;
            for j in range(Np):
                sxx = S[j]  ; syy = S[j+Np]; szz = -sxx-syy;
                sxy = S[j+xx]; syx = sxy;
                sxz = S[j+xx1]; szx = sxz;
                syz = S[j+xx2]; szy = syz;
                dx = rt[i]    - r[j]
                dy = rt[i+Nt] - r[j+Np]
                idr  = 1.0/sqrt( dx*dx + dy*dy);
                idr4 = idr*idr*idr*idr; idr6 = idr4*idr*idr; 
                fac1 = (H-rt[i+2*Nt])*(H-r[j+xx])*rt[i+2*Nt]*idr4

                srx = fac1*(sxx*dx    +  sxy*dy ); 
                sry = fac1*(sxy*dx    +  syy*dy );
                srr = fac1*(sxx*dx*dx + syy*dy*dy + 2*sxy*dx*dy )*idr*idr;
                
                vx += -2*srx + 4*srr*dx;
                vy += -2*sry + 4*srr*dy;
                    
            vv[i  ]  += mus*vx 
            vv[i+Nt] += mus*vy


    cpdef flowField3t(self, double [:] vv, double [:] rt, double [:] r, double [:] D, double H):
        """
        Compute flow field at field points due to 3t mode of th slip 
        ...

        Parameters
        ----------
        vv: np.array
            An array of flow at field points
            An array of size 3*Nt,
        rt: np.array
            An array of field points
            An array of size 3*Nt,
        r: np.array
            An array of positions
            An array of size 3*Np,
        D: np.array
            An array of 2s mode of the slip 
            An array of size r*Np,
        H: float 
            Height of the Hele-Shaw cell 
        """
        cdef int i, j, Np=self.Np, xx=2*Np, Nt=self.Nt
        cdef double dx, dy, dz, idr, idr2, idr5, Fdotidr, h2, hsq, tempF
        cdef double vx, vy, vz, tH = 2*H, Ddotidr
        cdef double mu = 1.0/(6*PI*self.eta*self.a), mu1 = mu*self.a*0.75, a2=self.a*self.a/3.0
        cdef double fac0 = 3/(PI*self.eta*H*H*H), fac1, fac2
 
        for i in prange(Nt, nogil=True):
            vx=0; vy=0; vz=0;
            for j in range(Np):
                dx = rt[i]   - r[j]
                dy = rt[i+Nt] - r[j+Np]
                idr = 1.0/sqrt( dx*dx + dy*dy)
                idr2=idr*idr
                Ddotidr = (D[j]*dx + D[j+Np]*dy)*idr2
                
                fac1 = fac0*(H-rt[i+2*Nt])*(H-r[j+xx])*rt[i+2*Nt]*r[j+xx]
                vx += fac1*(0.5*D[j]    - Ddotidr*dx)*idr2
                vy += fac1*(0.5*D[j+Np] - Ddotidr*dy)*idr2
                    
            vv[i  ]  += mu1*vx 
            vv[i+Nt] += mu1*vy
